from collections import defaultdict
from warnings import catch_warnings, simplefilter

import numpy as np


class Motif:
    def __init__(self, pattern):
        self.pattern = pattern
        self.indexes = defaultdict(list)
        self.average_occurrences = {}
        self.representative = None
        self.best_matches = {}
        self.naed = 0.0
        self.length = 0
        self._seglen = 0

    def __repr__(self):
        return f"Motif('{self.pattern}')"

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return other.pattern == self.pattern

    def record_index(self, i, j):
        """Record starting index of pattern in sequence i at position j."""
        self.indexes[i].append(j)
    
    def map(self, ts, seglen):
        """Map representative, matches, and naed using occurrences."""
        self._seglen = seglen
        self.length = len(self.pattern) * seglen
        self.set_average_occurrences(ts)
        self.set_representative()
        self.set_best_matches_and_naed(ts)

    def set_average_occurrences(self, ts):
        for i, indexes in self.indexes.items():
            occurrences = []
            for index in indexes:
                occurrence = self.get_occurrence(ts[i], index)
                occurrences.append(occurrence)
            with catch_warnings():
                simplefilter("ignore")
                self.average_occurrences[i] = np.nanmean(occurrences, axis=0)

    def get_occurrence(self, ts, index):
        start = index * self._seglen
        end = start + self.length

        # Ensure motif occurrences are all the same length
        if too_short := max(0, end - len(ts)):
            return np.hstack((ts[start : end], np.array(too_short * [np.nan])))
        return ts[start : end]

    def set_representative(self):
        self.representative = np.nanmean([ao for ao in self.average_occurrences.values()], axis=0)

    def set_best_matches_and_naed(self, ts):
        for i, indexes in self.indexes.items():
            best_match = 0
            min_naed = 10 ** 6

            for index in indexes:
                occurrence = self.get_occurrence(ts[i], index)
                naed = np.nansum((occurrence - self.representative) ** 2) ** 0.5
                if naed < min_naed:
                    min_naed = naed
                    best_match = index
            self.naed += min_naed
            self.best_matches[i] = best_match * self._seglen
        self.naed /= (len(self.indexes)) * (self.length)



"""Motif miner module.

This module defines the mine_motifs function, which takes a database of
time series and finds frequent or maximal motifs in it. The motifs can
be filtered for length and ranked using different strategies.
"""
from .preprocessing import standardise, sax
from .patterns import PatternMiner


class Miner:
    """Motif miner class.
    
    Parameters
    ----------
    minsup : float
        Fraction of time series a motif should occur in.
    seglen : int
        Segment length for Piecewise Aggregate Approximation.
    alphabet : int
        Alphabet size for discretisation.
    min_len : int
        Minimal pattern length.
    max_len : int
        Maximal pattern length.
        If 0, limit is infinity.
    max_overlap: float
        Maximal fraction of patterns contained in longer patters.
    k : int, optional
        Number of motifs to return.
        If 0, all motifs are returned.
    maximal : bool
        Return only maximal patterns. 
        If False, returns all frequent motifs.
    
    Attributes
    ----------
    motifs : list
        Constructed motifs ordered by the distances to their occurrences.
    """
    def __init__(
            self,
            minsup: float,
            seglen: int,
            alphabet: int,
            min_len: int = 3,
            max_len: int = 0,
            max_overlap: float = 0.9,
            k: int = 0,
    ):
        self.minsup = minsup
        self.seglen = seglen
        self.alphabet = alphabet
        self.min_len = min_len
        self.max_len = max_len
        self.max_overlap = max_overlap
        self.k = k

        self.motifs = None

    def mine(self, ts):
        """Perform all steps in motif mining pipeline.

        Parameters
        ----------
        ts : list
            Database of time series.

        Returns
        -------
        res: list
            frequent motifs.
        """
        standardised = standardise(ts)
        discretised = sax(standardised, self.seglen, self.alphabet)
        self.mine_patterns(discretised)
        self.map_patterns(standardised)
        self.sort_patterns()

        return self.motifs if not self.k else self.motifs[:self.k]

    def mine_patterns(self, ds):
        """Find frequent patterns in the sequences.

        Parameters
        ----------
        sequences : list
            Collection of time series discretised to sequences.
        """
        pm = PatternMiner(self.minsup, self.min_len, self.max_len, self.max_overlap)
        pm.mine(ds)
        self.motifs = list(pm.frequent.values())

    def map_patterns(self, ts):
        """Map patterns back to motifs."""
        for motif in self.motifs:
            motif.map(ts, self.seglen)

    def sort_patterns(self):
        """Sort patterns on their root mean squared error of representative and occurrences."""
        self.motifs.sort(key=lambda motif: motif.naed)




import unittest

from frm._frm_py.preprocessing import sax, standardise

from test_data import ts, norm, rag, seq_1, seq_2, rseq_1, rseq_2


class TestPreprocessing(unittest.TestCase):
    def test_standardise(self):
        for expected, got in zip(norm, standardise(ts)):
            for a, b in zip(expected, got):
                self.assertAlmostEqual(a, b, 5)
    
    def test_sax_seglen_1(self):
        got = sax(standardise(ts), 1, 3)
        self.assertEqual(seq_1, got)

    def test_rag_sax_seglen_1(self):
        got = sax(standardise(rag), 1, 3)
        self.assertEqual(rseq_1, got)

    def test_sax_seglen_2(self):
        got = sax(standardise(ts), 2, 3)
        self.assertEqual(seq_2, got)

    def test_rag_sax_seglen_2(self):
        got = sax(standardise(rag), 2, 3)
        self.assertEqual(rseq_2, got)



import unittest

from frm._frm_py.motif import Motif


class TestMotif(unittest.TestCase):
    def test_init(self):
        a = Motif(pattern='abc')
        self.assertEqual(a.pattern, 'abc')
    
    def test_record_index(self):
        a = Motif(pattern='abc')
        a.record_index(0, 1)
        self.assertEqual(list(a.indexes.keys()), [0])
        self.assertEqual(a.indexes[0], [1])



import unittest

from frm._frm_py.patterns import PatternMiner
from test_data import seq_1, rseq_1


class TestPatternMiner(unittest.TestCase):
    def test_lcs(self):
        pm = PatternMiner()

        self.assertEqual(pm.lcs('bbbbbbbbbb', 'bbbcbbb', 10, 7), 6)

    def test_mine(self):
        pm = PatternMiner(min_len=1, max_len=0, max_overlap=1)
        pm.mine(seq_1)

        expected = ['a', 'aa', 'c', 'ca', 'cc']
        
        self.assertListEqual(expected, sorted(pm.frequent.keys()))
    
    def test_rag_mine(self):
        pm = PatternMiner(min_len=1, max_overlap=1.1)
        pm.mine(rseq_1)

        expected = ['a', 'ab', 'abc', 'b', 'bc', 'c']
        
        self.assertListEqual(expected, sorted(pm.frequent.keys()))



"""Preprocessing module.

This module defines two time series preprocessing functions for standardisation and SAX representation.
"""
from typing import Union

from scipy.stats import zscore
from numpy import nan_to_num

# https://stackoverflow.com/a/60617044
numeric = Union[int, float]


def sax(timeseries: list[list[numeric]], seglen: int, alphabet: int) -> list[str]:
    """Symbolic Aggregate approXimation.
    
    Parameters
    ----------
    timeseries : list
        List of lists containing normalised time series to dicretise.
    seglen : int
        Segment length for PAA, factor by which discrete representation is shorter than time series.
    alphabet : int
        Alphabet size; number of discrete elements the time series are to be binned into.
    
    Returns
    -------
    List with a collection of discrete sequences from the time series.
    """
    return [get_sax(series, seglen, alphabet) for series in timeseries]


def get_sax(series: list[numeric], seglen: int, a: int) -> str:
    """Get SAX representation of one time series."""
    segments = [sum(series[i:i + seglen]) / seglen for i in range(0, len(series), seglen)]
    for i, seg in enumerate(segments):
        for b, c in breakpoints[a].items():
            if seg <= b:
                segments[i] = c
                break
        else:
            segments[i] = chr(ord(c) + 1)

    sax_representation = ''.join(segments)
    return sax_representation


def standardise(timeseries: list[list]):
    """Standardise time series.

    Standardises data to have a mean of zero and a standard deviation of one.

    Parameters
    ----------
    timeseries
        Database of time series to standardise.

    Returns
    -------
    standardised
        Database of standardised time series.
    """
    try:
        return nan_to_num(zscore(timeseries, axis=1))
    except ValueError:
        return [nan_to_num(zscore(ts)) for ts in timeseries]


# Defined in Lin, Keogh, Linardi, & Chiu (2003). A Symbolic Representation of Time Series,
# with Implications for Streaming Algorithms. Table 3.
breakpoints = {
    2: {0: 'a'},
    3: {-0.43: 'a', 0.43: 'b'},
    4: {-0.67: 'a', 0: 'b', 0.67: 'c'},
    5: {-0.84: 'a', -0.25: 'b', 0.25: 'c', 0.84: 'd'},
    6: {-0.97: 'a', -0.43: 'b', 0: 'c', 0.43: 'd', 0.97: 'e'},
    7: {-1.07: 'a', -0.57: 'b', -0.18: 'c', 0.18: 'd', 0.57: 'e', 1.07: 'f'},
    8: {-1.15: 'a', -0.67: 'b', -0.32: 'c', 0: 'd', 0.32: 'e', 0.67: 'f', 1.15: 'g'},
    9: {-1.22: 'a', -0.76: 'b', -0.43: 'c', -0.14: 'd', 0.14: 'e', 0.43: 'f', 0.76: 'g', 1.22: 'h'},
    10: {-1.28: 'a', -0.84: 'b', -0.52: 'c', -0.25: 'd', 0: 'e', 0.25: 'f', 0.52: 'g', 0.84: 'h', 1.28: 'i'},
}



"""Patterns module.

This module defines the PatternMiner class, a class that takes a collection of
sequences and mines frequent and maximal patterns from it. The patterns
can be filtered to have a certain minimum length.
"""
from .motif import Motif


class PatternMiner:
    """Mine patterns from a collection of sequences with PatternMiner.

    Parameters
    ----------
    minsup : float
        The minimum support for a pattern.
    
    Attributes
    ----------
    frequent : dict
        Dictionary of frequent motifs with string patterns as keys and Motif objects as values.
    """
    def __init__(
            self,
            minsup: float = 0.5,
            min_len: int = 3,
            max_len: int = 0,
            max_overlap: float = 0.9):
        self.minsup = minsup
        self.min_len = min_len
        self.max_len = max_len
        self.max_overlap = max_overlap

        self.frequent = {}

        # Frequency is easier to check than support
        self._min_freq = 0

        # Keep track of length we're currently mining
        self._k = 2

        # self.patterns[k] contains a list of patterns of length k
        self._patterns = [[], []]

    def mine(self, sequences):
        """Mine sequence motifs.

        Starting with patterns of length 1 (1-patterns), candidates are
        generated and checked for sufficient support. Then, k-patterns
        are generated based on (k-1)-patterns.

        sequences : list
            Collection of sequences with discrete values.
        """
        self._min_freq = len(sequences) * self.minsup

        # Mine 1-patterns separately from longer patterns
        self.mine_1_patterns(sequences)

        # If there were no frequent k-patterns, there can be no 
        # frequent (k+1)-patterns; stop
        while self._patterns[self._k - 1] and (not self.max_len or self._k <= self.max_len):
            self._patterns.append([])

            for candidate in self.get_candidates():
                # Count candidate occurrences
                self.frequent[candidate] = Motif(candidate)

                # Find candidate in sequences parent occurs in
                for seq, indexes in self.frequent[candidate[:-1]].indexes.items():
                    sequence = sequences[seq]
                    for index in indexes:
                        if sequence[index : index+self._k] == candidate:
                            self.frequent[candidate].record_index(seq, index)
                
                # Check if candidate complies with minimum support
                self.prune_infrequent(candidate)

            self._k += 1

        self.remove_redundant()
    
    def remove_redundant(self):
        """Remove redundant patterns.
        
        Not all frequent patterns that comply with the minimum support
        are interesting. For example, 1-patterns are too trivial to be
        considered interesting.
        Furthermore, mining only maximal patterns can leave patterns
        that are trivial because they are almost contained in other
        patterns. For instance, if aaaaaa is a frequent pattern, aabaa
        could be considered trivial, as almost all of its contents are
        contained in the longer pattern.

        Parameters
        ----------
        min_len : int, optional.
            Minimum length of a pattern to be considered interesting.
        o : float, optional.
            Maximal redundancy ratio, calculated by
            length(longest common subsequence) / length(shortest pattern).
            If the redundancy ratio exceeds `max_overlap`, the shortest pattern is
            pruned.
        """
        # Prune from frequent patterns
        self.frequent = {k: v for k, v in self.frequent.items() if len(k) >= self.min_len}

        # If max_overlap == 1, no overlap is too high
        if self.max_overlap == 1:
            return

        # Prune patterns for which x% is overlapping from frequent
        patterns = list(self.frequent)
        patterns.sort(key=lambda x: (len(x), x), reverse=True)
        pruned = []
        
        for p1 in patterns:
            if p1 in pruned:
                continue
            for p2 in patterns:
                n, m = len(p1), len(p2)
                # Only check unseen patterns and pairs
                if m > n or p1 == p2 or p2 in pruned:
                    continue

                # Check if shorter pattern consists mostly of lcs
                if self.lcs(p1, p2, n, m) / m > self.max_overlap:
                    self.frequent.pop(p2, 0)
                    pruned.append(p2)

    def lcs(self, p1: str, p2: str, n: int, m: int) -> int:
        """Longest common subsequence.
        
        Find the length of the longest sequence that is contained in two
        patterns. Uses dynamic programming algorithm adapted from
        https://www.geeksforgeeks.org/longest-common-subsequence-dp-4/

        Parameters
        ----------
        p1 : str
            String representation of first pattern.
        p2 : str
            String representation of second pattern.
        n : int
            Length of first pattern.
        m : int
            Length of second pattern.
        
        Returns
        -------
        res : int
            Length of the longest common subsequence of `p1` and `p2`.
        """
        # declaring the array for storing the dp values
        L = [[0]*(m+1) for i in range(n+1)]

        """Following steps build L[n+1][m+1] in bottom up fashion
        Note: L[i][j] contains length of LCS of X[0..i-1]
        and Y[0..j-1]"""
        for i in range(n+1):
            for j in range(m+1):
                if i == 0 or j == 0:
                    L[i][j] = 0
                elif p1[i-1] == p2[j-1]:
                    L[i][j] = L[i-1][j-1]+1
                else:
                    L[i][j] = max(L[i-1][j], L[i][j-1])

        # L[n][m]] contains the length of LCS of X[0..n-1] & Y[0..m-1]
        return L[n][m]
    
    def get_candidates(self) -> list[str]:
        """Use frequent (k-1)-patterns to generate candidate k-patterns.
        
        Candidates are generated by joining two patterns from L[k-1]
        if pattern_1[1:] == pattern_2[:-1].

        Returns
        -------
        candidates : list[str]
            Set of candidate patterns that have a length of k.
        """
        patterns = self._patterns[self._k - 1]
        return [p1 + p2[-1] for p1 in patterns for p2 in patterns if p1[1:] == p2[:-1]]
    
    def mine_1_patterns(self, sequences):
        """Mine frequent 1-patterns.
        
        One scan is made over the sequences to map 1-pattern indexes.

        sequences : list
            Collection of sequences with discrete values.
        """
        # Divide sequences into indexes
        for i, sequence in enumerate(sequences):
            for j, item in enumerate(sequence):
                if item not in self.frequent:
                    self.frequent[item] = Motif(item)
                self.frequent[item].record_index(i, j)

        # Prune infrequent patterns
        for a in list(self.frequent.keys()):
            self.prune_infrequent(a)
    
    def prune_infrequent(self, pattern: str):
        """Prune infrequent patterns.
        
        - Prunes patterns with a too low support;
        - Adds frequent patterns to list of frequent patterns

        Parameters
        ----------
        pattern : str
            The pattern that should be processed.
        """
        if len(self.frequent[pattern].indexes) < self._min_freq:
            # Delete indexes of this infrequent pattern
            self.frequent.pop(pattern)
        else:
            # Add to list of frequent k-patterns
            self._patterns[len(pattern)].append(pattern)



import unittest

from frm import Miner as CppMiner
from frm._frm_py.miner import Miner as PyMiner
from test_data import ts, rag, data


class TestMiner(unittest.TestCase):
    def test_frm_py_miner(self):
        miner = PyMiner(0.5, 1, 3, 1, 0, 1)
        motifs = miner.mine(ts)
        patterns = [m.pattern for m in motifs]
        self.assertListEqual(patterns, ['aa', 'a', 'ca', 'cc', 'c'])

    def test_rag_py_miner(self):
        miner = PyMiner(0.5, 1, 3, 1, 0, 1)
        motifs = miner.mine(rag)
        patterns = [m.pattern for m in motifs]
        self.assertListEqual(patterns, ['ab', 'abc', 'b', 'bc', 'a', 'c'])

    def test_frm_cpp_miner(self):
        miner = CppMiner(0.5, 1, 3, 1, 0, 1)
        motifs = miner.mine(ts)
        patterns = [m.pattern for m in motifs]
        self.assertListEqual(patterns, ['aa', 'a', 'ca', 'cc', 'c'])

    def test_rag_cpp_miner(self):
        miner = CppMiner(0.5, 1, 3, 1, 0, 1)
        motifs = miner.mine(rag)
        patterns = [m.pattern for m in motifs]
        self.assertListEqual(patterns, ['ab', 'abc', 'b', 'bc', 'a', 'c'])

    def test_equal_miners(self):
        py_miner = PyMiner(0.5, 1, 3, 1, 0, 1)
        py_motifs = py_miner.mine(ts)

        cpp_miner = CppMiner(0.5, 1, 3, 1, 0, 1)
        cpp_motifs = cpp_miner.mine(ts)

        for py_motif, cpp_motif in zip(py_motifs, cpp_motifs):
            self.assertAlmostEqual(py_motif.naed, cpp_motif.naed)

    def test_equal_rag_miners(self):
        py_miner = PyMiner(0.5, 1, 3, 1, 0, 1)
        py_motifs = py_miner.mine(rag)

        cpp_miner = CppMiner(0.5, 1, 3, 1, 0, 1)
        cpp_motifs = cpp_miner.mine(rag)

        for py_motif, cpp_motif in zip(py_motifs, cpp_motifs):
            self.assertAlmostEqual(py_motif.naed, cpp_motif.naed)

    def test_max_len_py_miner(self):
        miner = PyMiner(0.5, 1, 3, 1, 1, 1)
        motifs = miner.mine(ts)
        patterns = [m.pattern for m in motifs]
        self.assertListEqual(patterns, ['a', 'c'])

    def test_max_len_cpp_miner(self):
        miner = CppMiner(0.5, 1, 3, 1, 1, 1)
        motifs = miner.mine(ts)
        patterns = [m.pattern for m in motifs]
        self.assertListEqual(patterns, ['a', 'c'])

    def test_equal_long(self):
        seglen, alphabet = 2, 4
        py = PyMiner(0.5, seglen, alphabet)
        cpp = CppMiner(0.5, seglen, alphabet)

        py.mine(data)
        cpp.mine(data)

        for p, c in zip(py.motifs, cpp.motifs):
            self.assertEqual(p.pattern, c.pattern)
            self.assertAlmostEqual(p.naed, c.naed)
